# ruff: noqa: N802
import ast
import functools
import inspect
from collections.abc import Callable, Mapping
from contextlib import contextmanager
from types import FunctionType, MethodType
from typing import Any, Never

from sonolus.backend.excepthook import install_excepthook
from sonolus.backend.utils import get_function, scan_writes
from sonolus.script.debug import assert_true
from sonolus.script.internal.builtin_impls import BUILTIN_IMPLS
from sonolus.script.internal.context import Context, EmptyBinding, Scope, ValueBinding, ctx, set_ctx
from sonolus.script.internal.descriptor import SonolusDescriptor
from sonolus.script.internal.error import CompilationError
from sonolus.script.internal.impl import try_validate_value, validate_value
from sonolus.script.internal.value import Value
from sonolus.script.iterator import SonolusIterator
from sonolus.script.num import Num

_compiler_internal_ = True


def compile_and_call[**P, R](fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs) -> R:
    if not ctx():
        return fn(*args, **kwargs)
    return validate_value(generate_fn_impl(fn)(*args, **kwargs))


def generate_fn_impl(fn: Callable):
    install_excepthook()
    match fn:
        case Value() as value if value._is_py_():
            return generate_fn_impl(value._as_py_())
        case MethodType() as method:
            return functools.partial(generate_fn_impl(method.__func__), method.__self__)
        case FunctionType() as function:
            if getattr(function, "_meta_fn_", False):
                return function
            return functools.partial(eval_fn, function)
        case _:
            if callable(fn) and isinstance(fn, Value):
                return generate_fn_impl(fn.__call__)
            elif fn is type:
                return fn
            elif callable(fn):
                raise TypeError(f"Unsupported callable {fn!r}")
            else:
                raise TypeError(f"Not callable {fn!r}")


def eval_fn(fn: Callable, /, *args, **kwargs):
    source_file, node = get_function(fn)
    bound_args = inspect.signature(fn).bind(*args, **kwargs)
    bound_args.apply_defaults()
    closurevars = inspect.getclosurevars(fn)
    global_vars = {**closurevars.nonlocals, **closurevars.globals, **closurevars.builtins}
    return Visitor(source_file, bound_args, global_vars).run(node)


unary_ops = {
    ast.Invert: "__invert__",
    ast.UAdd: "__pos__",
    ast.USub: "__neg__",
}

bin_ops = {
    ast.Add: "__add__",
    ast.Sub: "__sub__",
    ast.Mult: "__mul__",
    ast.Div: "__truediv__",
    ast.FloorDiv: "__floordiv__",
    ast.Mod: "__mod__",
    ast.Pow: "__pow__",
    ast.LShift: "__lshift__",
    ast.RShift: "__rshift__",
    ast.BitOr: "__or__",
    ast.BitAnd: "__and__",
    ast.BitXor: "__xor__",
    ast.MatMult: "__matmul__",
}

rbin_ops = {
    ast.Add: "__radd__",
    ast.Sub: "__rsub__",
    ast.Mult: "__rmul__",
    ast.Div: "__rtruediv__",
    ast.FloorDiv: "__rfloordiv__",
    ast.Mod: "__rmod__",
    ast.Pow: "__rpow__",
    ast.LShift: "__rlshift__",
    ast.RShift: "__rrshift__",
    ast.BitOr: "__ror__",
    ast.BitAnd: "__rand__",
    ast.BitXor: "__rxor__",
    ast.MatMult: "__rmatmul__",
}

inplace_ops = {
    ast.Add: "__iadd__",
    ast.Sub: "__isub__",
    ast.Mult: "__imul__",
    ast.Div: "__itruediv__",
    ast.FloorDiv: "__ifloordiv__",
    ast.Mod: "__imod__",
    ast.Pow: "__ipow__",
    ast.LShift: "__ilshift__",
    ast.RShift: "__irshift__",
    ast.BitOr: "__ior__",
    ast.BitXor: "__ixor__",
    ast.BitAnd: "__iand__",
    ast.MatMult: "__imatmul__",
}

comp_ops = {
    ast.Eq: "__eq__",
    ast.NotEq: "__ne__",
    ast.Lt: "__lt__",
    ast.LtE: "__le__",
    ast.Gt: "__gt__",
    ast.GtE: "__ge__",
}

rcomp_ops = {
    ast.Eq: "__req__",
    ast.NotEq: "__rne__",
    ast.Lt: "__gt__",
    ast.LtE: "__ge__",
    ast.Gt: "__lt__",
    ast.GtE: "__le__",
    ast.In: "__contains__",
    ast.NotIn: "__contains__",
}


class Visitor(ast.NodeVisitor):
    source_file: str
    globals: dict[str, Any]
    bound_args: inspect.BoundArguments
    used_names: dict[str, int]
    return_ctxs: list[Context]  # Contexts at return statements, which will branch to the exit
    loop_head_ctxs: list[Context]  # Contexts at loop heads, from outer to inner
    break_ctxs: list[list[Context]]  # Contexts at break statements, from outer to inner

    def __init__(self, source_file: str, bound_args: inspect.BoundArguments, global_vars: dict[str, Any]):
        self.source_file = source_file
        self.globals = {}
        for k, v in global_vars.items():
            # Unfortunately, inspect.closurevars also includes attributes
            if v is ctx:
                raise ValueError("Unexpected use of ctx in non-meta function")
            value = try_validate_value(BUILTIN_IMPLS.get(id(v), v))
            if value is not None:
                self.globals[k] = value
        self.bound_args = bound_args
        self.used_names = {}
        self.return_ctxs = []
        self.loop_head_ctxs = []
        self.break_ctxs = []

    def run(self, node):
        before_ctx = ctx()
        set_ctx(before_ctx.branch_with_scope(None, Scope()))
        for name, value in self.bound_args.arguments.items():
            ctx().scope.set_value(name, validate_value(value))
        match node:
            case ast.FunctionDef(body=body):
                ctx().scope.set_value("$return", validate_value(None))
                for stmt in body:
                    self.visit(stmt)
            case _:
                raise NotImplementedError("Unsupported syntax")
        after_ctx = Context.meet([*self.return_ctxs, ctx()])
        result_binding = after_ctx.scope.get_binding("$return")
        if not isinstance(result_binding, ValueBinding):
            raise ValueError("Function has conflicting return values")
        set_ctx(after_ctx.branch_with_scope(None, before_ctx.scope.copy()))
        return result_binding.value

    def visit_FunctionDef(self, node):
        raise NotImplementedError("Nested functions are not supported")

    def visit_AsyncFunctionDef(self, node):
        raise NotImplementedError("Async functions are not supported")

    def visit_ClassDef(self, node):
        raise NotImplementedError("Classes within functions are not supported")

    def visit_Return(self, node):
        value = self.visit(node.value) if node.value else validate_value(None)
        ctx().scope.set_value("$return", value)
        self.return_ctxs.append(ctx())
        set_ctx(ctx().into_dead())

    def visit_Delete(self, node):
        raise NotImplementedError("Delete statements are not supported")

    def visit_Assign(self, node):
        value = self.visit(node.value)
        for target in node.targets:
            self.handle_assign(target, value)

    def visit_TypeAlias(self, node):
        raise NotImplementedError("Type aliases are not supported")

    def visit_AugAssign(self, node):
        lhs_value = self.visit(node.target)
        rhs_value = self.visit(node.value)
        inplace_fn_name = inplace_ops[type(node.op)]
        regular_fn_name = bin_ops[type(node.op)]
        right_fn_name = rbin_ops[type(node.op)]
        if hasattr(lhs_value, inplace_fn_name):
            result = self.handle_call(node, getattr(lhs_value, inplace_fn_name), rhs_value)
            if not self.is_not_implemented(result):
                if result is not lhs_value:
                    raise ValueError("Inplace operation must return the same object")
                # Skip the actual assignment because the inplace operation already did the job, as an optimization
                # There could be side effects of assignment, but that's atypical
                return
        if hasattr(lhs_value, regular_fn_name):
            result = self.handle_call(node, getattr(lhs_value, regular_fn_name), rhs_value)
            if not self.is_not_implemented(result):
                self.handle_assign(node.target, result)
                return
        if hasattr(rhs_value, right_fn_name):
            result = self.handle_call(node, getattr(rhs_value, right_fn_name), lhs_value)
            if not self.is_not_implemented(result):
                self.handle_assign(node.target, result)
                return
        raise NotImplementedError("Unsupported augmented assignment")

    def visit_AnnAssign(self, node):
        value = self.visit(node.value)
        self.handle_assign(node.target, value)

    def visit_For(self, node):
        iterator = iter(self.visit(node.iter))
        if not isinstance(iterator, SonolusIterator):
            raise ValueError("Unsupported iterator")
        writes = scan_writes(node)
        header_ctx = ctx().prepare_loop_header(writes)
        self.loop_head_ctxs.append(header_ctx)
        self.break_ctxs.append([])
        set_ctx(header_ctx)
        has_next = self.ensure_boolean_num(self.handle_call(node, iterator.has_next))
        ctx().test = has_next.ir()
        body_ctx = ctx().branch(None)
        else_ctx = ctx().branch(0)

        set_ctx(body_ctx)
        self.handle_assign(node.target, self.handle_call(node, iterator.next))
        for stmt in node.body:
            self.visit(stmt)
        ctx().branch_to_loop_header(header_ctx)

        set_ctx(else_ctx)
        for stmt in node.orelse:
            self.visit(stmt)
        else_end_ctx = ctx()

        break_ctxs = self.break_ctxs.pop()
        after_ctx = Context.meet([else_end_ctx, *break_ctxs])
        set_ctx(after_ctx)

    def visit_While(self, node):
        writes = scan_writes(node)
        header_ctx = ctx().prepare_loop_header(writes)
        self.loop_head_ctxs.append(header_ctx)
        self.break_ctxs.append([])
        set_ctx(header_ctx)
        test = self.ensure_boolean_num(self.visit(node.test))
        ctx().test = test.ir()
        body_ctx = ctx().branch(None)
        else_ctx = ctx().branch(0)

        set_ctx(body_ctx)
        for stmt in node.body:
            self.visit(stmt)
        ctx().branch_to_loop_header(header_ctx)

        set_ctx(else_ctx)
        for stmt in node.orelse:
            self.visit(stmt)
        else_end_ctx = ctx()

        break_ctxs = self.break_ctxs.pop()
        after_ctx = Context.meet([else_end_ctx, *break_ctxs])
        set_ctx(after_ctx)

    def visit_If(self, node):
        test = self.ensure_boolean_num(self.visit(node.test))

        if test._is_py_():
            if test._as_py_():
                for stmt in node.body:
                    self.visit(stmt)
            else:
                for stmt in node.orelse:
                    self.visit(stmt)
            return

        ctx_init = ctx()
        ctx_init.test = test.ir()
        true_ctx = ctx_init.branch(None)
        false_ctx = ctx_init.branch(0)

        set_ctx(true_ctx)
        for stmt in node.body:
            self.visit(stmt)
        true_end_ctx = ctx()

        set_ctx(false_ctx)
        for stmt in node.orelse:
            self.visit(stmt)
        false_end_ctx = ctx()

        set_ctx(Context.meet([true_end_ctx, false_end_ctx]))

    def visit_With(self, node):
        raise NotImplementedError("With statements are not supported")

    def visit_AsyncWith(self, node):
        raise NotImplementedError("Async with statements are not supported")

    def visit_Match(self, node):
        subject = self.visit(node.subject)
        end_ctxs = []
        for case in node.cases:
            if not ctx().live:
                break
            true_ctx, false_ctx = self.handle_match_pattern(subject, case.pattern)
            if not true_ctx.live:
                set_ctx(false_ctx)
                continue
            set_ctx(true_ctx)
            guard = self.ensure_boolean_num(self.visit(case.guard)) if case.guard else validate_value(True)
            if guard._is_py_():
                if guard._as_py_():
                    for stmt in case.body:
                        self.visit(stmt)
                    end_ctxs.append(ctx())
                else:
                    end_ctxs.append(ctx())
            else:
                ctx().test = guard.ir()
                true_ctx = ctx().branch(None)
                false_ctx = ctx().branch(0)
                set_ctx(true_ctx)
                for stmt in case.body:
                    self.visit(stmt)
                end_ctxs.append(ctx())
            set_ctx(false_ctx)
        if end_ctxs:
            set_ctx(Context.meet(end_ctxs))

    def handle_match_pattern(self, subject: Value, pattern: ast.pattern) -> tuple[Context, Context]:
        match pattern:
            case ast.MatchValue(value=value):
                value = self.visit(value)
                test = self.ensure_boolean_num(subject == value)
                ctx_init = ctx()
                ctx_init.test = test.ir()
                true_ctx = ctx_init.branch(None)
                false_ctx = ctx_init.branch(0)
                return true_ctx, false_ctx
            case ast.MatchSingleton(value=value):
                match value:
                    case True:
                        test = self.ensure_boolean_num(subject)
                    case False:
                        test = self.ensure_boolean_num(subject).not_()
                    case None:
                        test = Num._accept_(subject._is_py_() and subject._as_py_() is None)
                    case _:
                        raise NotImplementedError("Unsupported match singleton")
                ctx_init = ctx()
                ctx_init.test = test.ir()
                true_ctx = ctx_init.branch(None)
                false_ctx = ctx_init.branch(0)
                return true_ctx, false_ctx
            case ast.MatchSequence():
                raise NotImplementedError("Match sequences are not supported")
            case ast.MatchMapping():
                raise NotImplementedError("Match mappings are not supported")
            case ast.MatchClass(cls=cls, patterns=patterns, kwd_attrs=kwd_attrs, kwd_patterns=kwd_patterns):
                from sonolus.script.comptime import Comptime
                from sonolus.script.internal.generic import validate_type_spec

                cls = validate_type_spec(self.visit(cls))
                if not isinstance(cls, type):
                    raise TypeError("Class is not a type")
                if issubclass(cls, Comptime):
                    raise TypeError("Comptime is not supported in match patterns")
                if not isinstance(subject, cls):
                    return ctx().into_dead(), ctx()
                if patterns:
                    if not hasattr(cls, "__match_args__"):
                        raise TypeError("Class does not support match patterns")
                    if len(cls.__match_args__) < len(patterns):
                        raise ValueError("Too many match patterns")
                    # kwd_attrs can't be mixed with patterns on the syntax level,
                    # so we can just set it like this since it's empty
                    kwd_attrs = cls.__match_args__[: len(patterns)]
                    kwd_patterns = patterns
                if kwd_attrs:
                    true_ctx = ctx()
                    false_ctxs = []
                    for attr, subpattern in zip(kwd_attrs, kwd_patterns, strict=False):
                        if not hasattr(subject, attr):
                            raise AttributeError(f"Object has no attribute {attr}")
                        value = self.handle_getattr(subpattern, subject, attr)
                        true_ctx, false_ctx = self.handle_match_pattern(value, subpattern)
                        false_ctxs.append(false_ctx)
                        set_ctx(true_ctx)
                    return true_ctx, Context.meet(false_ctxs)
                return ctx(), ctx().into_dead()
            case ast.MatchStar():
                raise NotImplementedError("Match stars are not supported")
            case ast.MatchAs(pattern=pattern, name=name):
                if pattern:
                    true_ctx, false_ctx = self.handle_match_pattern(subject, pattern)
                    if name:
                        true_ctx.scope.set_value(name, subject)
                    return true_ctx, false_ctx
                else:
                    if name:
                        ctx().scope.set_value(name, subject)
                    return ctx(), ctx().into_dead()
            case ast.MatchOr():
                true_ctxs = []
                false_ctx = ctx()
                assert pattern.patterns
                for subpattern in pattern.patterns:
                    true_ctx, false_ctx = self.handle_match_pattern(subject, subpattern)
                    true_ctxs.append(true_ctx)
                    set_ctx(false_ctx)
                return Context.meet(true_ctxs), false_ctx

    def visit_Raise(self, node):
        raise NotImplementedError("Raise statements are not supported")

    def visit_Try(self, node):
        raise NotImplementedError("Try statements are not supported")

    def visit_TryStar(self, node):
        raise NotImplementedError("Try* statements are not supported")

    def visit_Assert(self, node):
        self.handle_call(
            node, assert_true, self.visit(node.test), self.visit(node.msg) if node.msg else validate_value(None)
        )

    def visit_Import(self, node):
        raise NotImplementedError("Import statements are not supported")

    def visit_ImportFrom(self, node):
        raise NotImplementedError("Import statements are not supported")

    def visit_Global(self, node):
        raise NotImplementedError("Global statements are not supported")

    def visit_Nonlocal(self, node):
        raise NotImplementedError("Nonlocal statements are not supported")

    def visit_Expr(self, node):
        return self.visit(node.value)

    def visit_Pass(self, node):
        pass

    def visit_Break(self, node):
        self.break_ctxs[-1].append(ctx())
        set_ctx(ctx().into_dead())

    def visit_Continue(self, node):
        ctx().branch_to_loop_header(self.loop_head_ctxs[-1])
        set_ctx(ctx().into_dead())

    def visit_BoolOp(self, node) -> Value:
        match node.op:
            case ast.And():
                handler = self.handle_and
            case ast.Or():
                handler = self.handle_or
            case _:
                raise NotImplementedError(f"Unsupported bool operator {node.op}")

        if not node.values:
            raise ValueError("Bool operator requires at least one operand")
        if len(node.values) == 1:
            return self.visit(node.values[0])
        initial, *rest = node.values
        return handler(self.visit(initial), ast.copy_location(ast.BoolOp(op=node.op, values=rest), node))

    def visit_NamedExpr(self, node):
        value = self.visit(node.value)
        self.handle_assign(node.target, value)
        return value

    def visit_BinOp(self, node):
        lhs = self.visit(node.left)
        rhs = self.visit(node.right)
        op = bin_ops[type(node.op)]
        if hasattr(lhs, op):
            result = self.handle_call(node, getattr(lhs, op), rhs)
            if not self.is_not_implemented(result):
                return result
        if hasattr(rhs, rbin_ops[type(node.op)]):
            result = self.handle_call(node, getattr(rhs, rbin_ops[type(node.op)]), lhs)
            if not self.is_not_implemented(result):
                return result
        raise NotImplementedError(f"Unsupported operand types for binary operator {node.op}")

    def visit_UnaryOp(self, node):
        operand = self.visit(node.operand)
        if isinstance(node.op, ast.Not):
            return self.ensure_boolean_num(operand).not_()
        op = unary_ops[type(node.op)]
        if hasattr(operand, op):
            return self.handle_call(node, getattr(operand, op))
        raise NotImplementedError(f"Unsupported operand type for unary operator {node.op}")

    def visit_Lambda(self, node):
        raise NotImplementedError("Lambda functions are not supported")

    def visit_IfExp(self, node):
        test = self.ensure_boolean_num(self.visit(node.test))

        if test._is_py_():
            if test._as_py_():
                return self.visit(node.body)
            else:
                return self.visit(node.orelse)

        res_name = self.new_name("ifexp")
        ctx_init = ctx()
        ctx_init.test = test.ir()

        set_ctx(ctx_init.branch(None))
        true_value = self.visit(node.body)
        ctx().scope.set_value(res_name, true_value)
        ctx_true = ctx()

        set_ctx(ctx_init.branch(0))
        false_value = self.visit(node.orelse)
        ctx().scope.set_value(res_name, false_value)
        ctx_false = ctx()

        set_ctx(Context.meet([ctx_true, ctx_false]))
        return ctx().scope.get_value(res_name)

    def visit_Dict(self, node):
        return validate_value({self.visit(k): self.visit(v) for k, v in zip(node.keys, node.values, strict=True)})

    def visit_Set(self, node):
        raise NotImplementedError("Set literals are not supported")

    def visit_ListComp(self, node):
        raise NotImplementedError("List comprehensions are not supported")

    def visit_SetComp(self, node):
        raise NotImplementedError("Set comprehensions are not supported")

    def visit_DictComp(self, node):
        raise NotImplementedError("Dict comprehensions are not supported")

    def visit_GeneratorExp(self, node):
        raise NotImplementedError("Generator expressions are not supported")

    def visit_Await(self, node):
        raise NotImplementedError("Await expressions are not supported")

    def visit_Yield(self, node):
        raise NotImplementedError("Yield expressions are not supported")

    def visit_YieldFrom(self, node):
        raise NotImplementedError("Yield from expressions are not supported")

    def visit_Compare(self, node):
        result_name = self.new_name("compare")
        ctx().scope.set_value(result_name, Num._accept_(0))
        l_val = self.visit(node.left)
        false_ctxs = []
        for i, (op, rhs) in enumerate(zip(node.ops, node.comparators, strict=True)):
            r_val = self.visit(rhs)
            inverted = isinstance(op, ast.NotIn)
            result = None
            if isinstance(op, ast.Is | ast.IsNot):
                if not (r_val._is_py_() and r_val._as_py_() is None):
                    raise TypeError("The right operand of 'is' must be None")
                if isinstance(op, ast.Is):
                    result = Num._accept_(l_val._is_py_() and l_val._as_py_() is None)
                else:
                    result = Num._accept_(not (l_val._is_py_() and l_val._as_py_() is None))
            elif type(op) in comp_ops and hasattr(l_val, comp_ops[type(op)]):
                result = self.handle_call(node, getattr(l_val, comp_ops[type(op)]), r_val)
            if (
                (result is None or self.is_not_implemented(result))
                and type(op) in rcomp_ops
                and hasattr(r_val, rcomp_ops[type(op)])
            ):
                result = self.handle_call(node, getattr(r_val, rcomp_ops[type(op)]), l_val)
            if result is None or self.is_not_implemented(result):
                raise NotImplementedError(f"Unsupported comparison operator {op}")
            result = self.ensure_boolean_num(result)
            if inverted:
                result = result.not_()
            curr_ctx = ctx()
            if i == len(node.ops) - 1:
                curr_ctx.scope.set_value(result_name, result)
            else:
                curr_ctx.test = result.ir()
                true_ctx = curr_ctx.branch(None)
                false_ctx = curr_ctx.branch(0)
                false_ctxs.append(false_ctx)
                set_ctx(true_ctx)
                l_val = r_val
        last_ctx = ctx()  # This is the result of the last comparison returning true
        set_ctx(Context.meet([last_ctx, *false_ctxs]))
        return ctx().scope.get_value(result_name)

    def visit_Call(self, node):
        fn = self.visit(node.func)
        args = []
        kwargs = {}
        for arg in node.args:
            if isinstance(arg, ast.Starred):
                args.extend(self.handle_starred(self.visit(arg.value)))
            else:
                args.append(self.visit(arg))
        for keyword in node.keywords:
            if keyword.arg:
                kwargs[keyword.arg] = self.visit(keyword.value)
            else:
                value = self.visit(keyword.value)
                if value._is_py_() and isinstance(value._as_py_(), Mapping):
                    kwargs.update(value._as_py_())
                else:
                    raise ValueError("Starred keyword arguments (**kwargs) must be dictionaries")
        return self.handle_call(node, fn, *args, **kwargs)

    def visit_FormattedValue(self, node):
        raise NotImplementedError("F-strings are not supported")

    def visit_JoinedStr(self, node):
        raise NotImplementedError("F-strings are not supported")

    def visit_Constant(self, node):
        return validate_value(node.value)

    def visit_Attribute(self, node):
        return self.handle_getattr(node, self.visit(node.value), node.attr)

    def visit_Subscript(self, node):
        value = self.visit(node.value)
        slice_value = self.visit(node.slice)
        return self.handle_getitem(node, value, slice_value)

    def visit_Starred(self, node):
        raise NotImplementedError("Starred expressions are not supported")

    def visit_Name(self, node):
        if isinstance(ctx().scope.get_binding(node.id), EmptyBinding) and node.id in self.globals:
            # globals can have false positives due to limitations of inspect.closurevars
            # so we need to check that it's not defined as a local variable
            return self.globals[node.id]
        return ctx().scope.get_value(node.id)

    def visit_List(self, node):
        raise NotImplementedError("List literals are not supported")

    def visit_Tuple(self, node):
        values = []
        for elt in node.elts:
            if isinstance(elt, ast.Starred):
                values.extend(self.handle_starred(self.visit(elt.value)))
            else:
                values.append(self.visit(elt))
        return validate_value(tuple(values))

    def visit_Slice(self, node):
        raise NotImplementedError("Slices are not supported")

    def handle_assign(self, target: ast.stmt | ast.expr, value: Value):
        match target:
            case ast.Name(id=name):
                ctx().scope.set_value(name, value)
            case ast.Attribute(value=attr_value, attr=attr):
                attr_value = self.visit(attr_value)
                self.handle_setattr(target, attr_value, attr, value)
            case ast.Subscript(value=sub_value, slice=slice_expr):
                sub_value = self.visit(sub_value)
                slice_value = self.visit(slice_expr)
                self.handle_setitem(target, sub_value, slice_value, value)
            case ast.Tuple(elts=elts) | ast.List(elts=elts):
                values = self.handle_starred(value)
                if len(elts) != len(values):
                    raise ValueError("Unpacking assignment requires the same number of elements")
                for elt, v in zip(elts, values, strict=False):
                    self.handle_assign(elt, validate_value(v))
            case ast.Starred():
                raise NotImplementedError("Starred assignment is not supported")
            case _:
                raise NotImplementedError("Unsupported assignment target")

    def handle_and(self, l_val: Value, r_expr: ast.expr) -> Value:
        ctx_init = ctx()
        l_val = self.ensure_boolean_num(l_val)
        ctx_init.test = l_val.ir()
        res_name = self.new_name("and")

        set_ctx(ctx_init.branch(None))
        r_val = self.ensure_boolean_num(self.visit(r_expr))
        ctx().scope.set_value(res_name, r_val)
        ctx_true = ctx()

        set_ctx(ctx_init.branch(0))
        ctx().scope.set_value(res_name, Num._accept_(0))
        ctx_false = ctx()

        set_ctx(Context.meet([ctx_true, ctx_false]))
        if l_val._is_py_() and r_val._is_py_():
            return Num._accept_(l_val._as_py_() and r_val._as_py_())
        return ctx().scope.get_value(res_name)

    def handle_or(self, l_val: Value, r_expr: ast.expr) -> Value:
        ctx_init = ctx()
        l_val = self.ensure_boolean_num(l_val)
        ctx_init.test = l_val.ir()
        res_name = self.new_name("or")

        set_ctx(ctx_init.branch(None))
        ctx().scope.set_value(res_name, l_val)
        ctx_true = ctx()

        set_ctx(ctx_init.branch(0))
        r_val = self.ensure_boolean_num(self.visit(r_expr))
        ctx().scope.set_value(res_name, r_val)
        ctx_false = ctx()

        set_ctx(Context.meet([ctx_true, ctx_false]))
        if l_val._is_py_() and r_val._is_py_():
            return Num._accept_(l_val._as_py_() or r_val._as_py_())
        return ctx().scope.get_value(res_name)

    def generic_visit(self, node):
        if isinstance(node, ast.stmt | ast.expr):
            with self.reporting_errors_at_node(node):
                raise NotImplementedError(f"Unsupported syntax: {type(node).__name__}")
        raise NotImplementedError(f"Unsupported syntax: {type(node).__name__}")

    def handle_getattr(self, node: ast.stmt | ast.expr, target: Value, key: str) -> Value:
        with self.reporting_errors_at_node(node):
            if target._is_py_():
                target = target._as_py_()
            descriptor = type(target).__dict__.get(key)
            match descriptor:
                case property(fget=getter):
                    return self.handle_call(node, getter, target)
                case SonolusDescriptor() | FunctionType() | classmethod() | staticmethod() | None:
                    return validate_value(getattr(target, key))
                case non_descriptor if not hasattr(non_descriptor, "__get__"):
                    return validate_value(getattr(target, key))
                case _:
                    raise TypeError(f"Unsupported field or descriptor {key}")

    def handle_setattr(self, node: ast.stmt | ast.expr, target: Value, key: str, value: Value):
        with self.reporting_errors_at_node(node):
            if target._is_py_():
                target = target._as_py_()
            descriptor = getattr(type(target), key, None)
            match descriptor:
                case property(fset=setter):
                    if setter is None:
                        raise AttributeError(f"Cannot set attribute {key} because property has no setter")
                    self.handle_call(node, setter, target, value)
                case SonolusDescriptor():
                    setattr(target, key, value)
                case _:
                    raise TypeError(f"Unsupported field or descriptor {key}")

    def handle_call[**P, R](
        self, node: ast.stmt | ast.expr, fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs
    ) -> R:
        """Handles a call to the given callable."""
        if (
            isinstance(fn, Value)
            and fn._is_py_()
            and isinstance(fn._as_py_(), type)
            and issubclass(fn._as_py_(), Value)
        ):
            return validate_value(self.execute_at_node(node, fn._as_py_(), *args, **kwargs))
        else:
            return self.execute_at_node(node, lambda: validate_value(compile_and_call(fn, *args, **kwargs)))

    def handle_getitem(self, node: ast.stmt | ast.expr, target: Value, key: Value) -> Value:
        with self.reporting_errors_at_node(node):
            if target._is_py_():
                target = target._as_py_()
                if key._is_py_():
                    return validate_value(target[key._as_py_()])
                if isinstance(target, Value) and hasattr(target, "__getitem__"):
                    return self.handle_call(node, target.__getitem__, key)
                raise TypeError(f"Cannot get items on {type(target).__name__}")
            else:
                if isinstance(target, Value) and hasattr(target, "__getitem__"):
                    return self.handle_call(node, target.__getitem__, key)
                raise TypeError(f"Cannot get items on {type(target).__name__}")

    def handle_setitem(self, node: ast.stmt | ast.expr, target: Value, key: Value, value: Value):
        with self.reporting_errors_at_node(node):
            if target._is_py_():
                target = target._as_py_()
                if key._is_py_():
                    target[key._as_py_()] = value._as_py_()
                if isinstance(target, Value) and hasattr(target, "__setitem__"):
                    return self.handle_call(node, target.__setitem__, key, value)
                raise TypeError(f"Cannot set items on {type(target).__name__}")
            else:
                if isinstance(target, Value) and hasattr(target, "__setitem__"):
                    return self.handle_call(node, target.__setitem__, key, value)
                raise TypeError(f"Cannot set items on {type(target).__name__}")

    def handle_starred(self, value: Value) -> tuple[Value, ...]:
        if value._is_py_() and isinstance(value._as_py_(), tuple):
            return value._as_py_()
        raise ValueError("Unsupported starred expression")

    def is_not_implemented(self, value):
        value = validate_value(value)
        return value._is_py_() and value._as_py_() is NotImplemented

    def ensure_boolean_num(self, value) -> Num:
        # This just checks the type for now, although we could support custom __bool__ implementations in the future
        if not isinstance(value, Num):
            raise TypeError(f"Invalid type where a bool (Num) was expected: {type(value).__name__}")
        return value

    def raise_exception_at_node(self, node: ast.stmt | ast.expr, cause: Exception) -> Never:
        """Throws a compilation error at the given node."""

        def thrower() -> Never:
            raise CompilationError(str(cause)) from cause

        self.execute_at_node(node, thrower)

    def execute_at_node[**P, R](
        self, node: ast.stmt | ast.expr, fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs
    ) -> R:
        """Executes the given function at the given node for a better traceback."""
        expr = ast.Expression(
            body=ast.Call(
                func=ast.Name(id="fn", ctx=ast.Load()),
                args=[ast.Starred(value=ast.Name(id="args", ctx=ast.Load()), ctx=ast.Load())],
                keywords=[ast.keyword(value=ast.Name(id="kwargs", ctx=ast.Load()), arg=None)],
                lineno=node.lineno,
                col_offset=node.col_offset,
                end_lineno=node.end_lineno,
                end_col_offset=node.end_col_offset,
            ),
        )
        expr = ast.fix_missing_locations(expr)
        return eval(
            compile(expr, filename=self.source_file, mode="eval"),
            {"fn": fn, "args": args, "kwargs": kwargs, "_filter_traceback_": True},
        )

    @contextmanager
    def reporting_errors_at_node(self, node: ast.stmt | ast.expr):
        try:
            yield
        except CompilationError as e:
            raise e from None
        except Exception as e:
            self.raise_exception_at_node(node, e)

    def new_name(self, name: str):
        self.used_names[name] = self.used_names.get(name, 0) + 1
        return f"${name}_{self.used_names[name]}"



from __future__ import annotations

from sonolus.script.array import Array
from sonolus.script.debug import error
from sonolus.script.iterator import ArrayLike, SonolusIterator
from sonolus.script.range import Range
from sonolus.script.record import Record
from sonolus.script.values import alloc, copy


class Pair[T, U](Record):
    first: T
    second: U


class VarArray[T, Capacity](Record, ArrayLike[T]):
    _size: int
    _array: Array[T, Capacity]

    @classmethod
    def new(cls):
        element_type = cls._get_type_arg_(T)
        capacity = cls._get_type_arg_(Capacity)
        return cls(0, alloc(Array[element_type, capacity]))

    def size(self) -> int:
        return self._size

    @classmethod
    def capacity(cls) -> int:
        return cls._get_type_arg_(Capacity)

    def is_full(self) -> bool:
        return self._size == self.capacity()

    def __getitem__(self, item) -> T:
        return self._array[item]

    def __setitem__(self, key: int, value: T):
        self._array[key] = value

    def append(self, value: T):
        """Appends a copy of the given value to the end of the array."""
        assert self._size < self._array.size()
        self._array[self._size] = value
        self._size += 1

    def extend(self, values: ArrayLike[T]):
        """Appends copies of the values in the given array to the end of the array."""
        for value in values:
            self.append(value)

    def pop(self, index: int | None = None) -> T:
        """Removes and returns a copy of the value at the given index.

        Preserves the relative order of the elements.
        """
        if index is None:
            index = self._size - 1
        assert 0 <= index < self._size
        value = copy(self._array[index])
        self._size -= 1
        if index < self._size:
            for i in Range(index, self._size):
                self._array[i] = self._array[i + 1]
        return value

    def insert(self, index: int, value: T):
        """Inserts a copy of the given value at the given index.

        Preserves the relative order of the elements.
        """
        assert 0 <= index <= self._size
        assert self._size < self._array.size()
        self._size += 1
        for i in Range(self._size - 1, index, -1):
            self._array[i] = self._array[i - 1]
        self._array[index] = value

    def remove(self, value: T) -> bool:
        """Removes the first occurrence of the given value, returning whether the value was removed.

        Preserves the relative order of the elements.
        """
        index = self.index_of(value)
        if index < 0:
            return False
        self.pop(index)
        return True

    def clear(self):
        """Sets size to zero."""
        self._size = 0

    def set_add(self, value: T) -> bool:
        """Adds a copy of the given value if it is not already present, returning whether the value was added.

        If the value is already present, the array is not modified.
        If the array is full, the value is not added.
        """
        if self._size >= self._array.size():
            return False
        if value in self:
            return False
        self.append(value)
        return True

    def set_remove(self, value: T) -> bool:
        """Removes the first occurrence of the given value, returning whether the value was removed.

        Does not preserve the relative order of the elements.
        """
        index = self.index_of(value)
        if index < 0:
            return False
        if index < self._size - 1:
            self._array[index] = self._array[self._size - 1]
        self._size -= 1
        return True

    def __eq__(self, other):
        if self.size() != other.size():
            return False
        i = 0
        while i < self.size():
            if self[i] != other[i]:
                return False
            i += 1
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        raise TypeError("unhashable type: 'VarArray'")


class ArrayMapEntry[K, V](Record):
    key: K
    value: V


class ArrayMap[K, V, Capacity](Record):
    _size: int
    _array: Array[ArrayMapEntry[K, V], Capacity]

    @classmethod
    def new(cls):
        key_type = cls._get_type_arg_(K)
        value_type = cls._get_type_arg_(V)
        capacity = cls._get_type_arg_(Capacity)
        return cls(0, alloc(Array[ArrayMapEntry[key_type, value_type], capacity]))

    def size(self) -> int:
        return self._size

    @classmethod
    def capacity(cls) -> int:
        return cls._get_type_arg_(Capacity)

    def is_full(self) -> bool:
        return self._size == self.capacity()

    def keys(self) -> SonolusIterator[K]:
        return _ArrayMapKeyIterator(self, 0)

    def values(self) -> SonolusIterator[V]:
        return ArrayMapValueIterator(self, 0)

    def items(self) -> SonolusIterator[tuple[K, V]]:
        return ArrayMapEntryIterator(self, 0)

    def __getitem__(self, key: K) -> V:
        for i in Range(self._size):
            entry = self._array[i]
            if entry.key == key:
                return entry.value
        error()

    def __setitem__(self, key: K, value: V):
        for i in Range(self._size):
            entry = self._array[i]
            if entry.key == key:
                entry.value = value
                return
        # assert self._size < self.capacity()
        self._array[self._size] = ArrayMapEntry(key, value)
        self._size += 1

    def __contains__(self, key: K) -> bool:
        for i in Range(self._size):  # noqa: SIM110
            if self._array[i].key == key:
                return True
        return False

    def pop(self, key: K) -> V:
        for i in Range(self._size):
            entry = self._array[i]
            if entry.key == key:
                value = copy(entry.value)
                self._size -= 1
                if i < self._size:
                    self._array[i] = self._array[self._size]
                return value
        error()

    def clear(self):
        self._size = 0


class _ArrayMapKeyIterator[K, V, Capacity](Record, SonolusIterator):
    _map: ArrayMap[K, V, Capacity]
    _index: int

    def has_next(self) -> bool:
        return self._index < self._map.size()

    def next(self) -> K:
        key = self._map._array[self._index].key
        self._index += 1
        return key


class ArrayMapValueIterator[K, V, Capacity](Record, SonolusIterator):
    _map: ArrayMap[K, V, Capacity]
    _index: int

    def has_next(self) -> bool:
        return self._index < self._map.size()

    def next(self) -> V:
        value = self._map._array[self._index].value
        self._index += 1
        return value


class ArrayMapEntryIterator[K, V, Capacity](Record, SonolusIterator):
    _map: ArrayMap[K, V, Capacity]
    _index: int

    def has_next(self) -> bool:
        return self._index < self._map.size()

    def next(self) -> tuple[K, V]:
        entry = self._map._array[self._index]
        self._index += 1
        return entry.key, entry.value



from sonolus.backend.flow import BasicBlock
from sonolus.backend.passes import CompilerPass


class CoalesceFlow(CompilerPass):
    def run(self, entry: BasicBlock) -> BasicBlock:
        queue = [entry]
        processed = set()
        while queue:
            block = queue.pop()
            if block in processed:
                continue
            processed.add(block)
            if len(block.outgoing) != 1:
                queue.extend(edge.dst for edge in block.outgoing)
                continue
            next_block = next(iter(block.outgoing)).dst
            if len(next_block.incoming) != 1:
                queue.append(next_block)
                continue
            block.statements.extend(next_block.statements)
            block.test = next_block.test
            for edge in next_block.outgoing:
                edge.src = block
            block.outgoing = next_block.outgoing
            processed.add(next_block)
            queue.extend(edge.dst for edge in block.outgoing)
            processed.remove(block)
            queue.append(block)
        return entry



from typing import Self

from sonolus.backend.ops import Op
from sonolus.script.debug import error
from sonolus.script.internal.native import native_function
from sonolus.script.record import Record


class Interval(Record):
    """A closed interval."""

    start: float
    end: float

    @property
    def length(self) -> float:
        return self.end - self.start

    @property
    def is_empty(self) -> bool:
        return self.start > self.end

    @property
    def mid(self) -> float:
        return (self.start + self.end) / 2

    @property
    def tuple(self):
        return self.start, self.end

    def __contains__(self, item: Self | float | int) -> bool:
        match item:
            case Interval(start, end):
                return self.start <= start and end <= self.end
            case float() | int() as value:
                return self.start <= value <= self.end
            case _:
                error("Invalid type for interval check")

    def __add__(self, other: float | int) -> Self:
        return Interval(self.start + other, self.end + other)

    def __sub__(self, other: float | int) -> Self:
        return Interval(self.start - other, self.end - other)

    def __mul__(self, other: float | int) -> Self:
        return Interval(self.start * other, self.end * other)

    def __truediv__(self, other: float | int) -> Self:
        return Interval(self.start / other, self.end / other)

    def __floordiv__(self, other: float | int) -> Self:
        return Interval(self.start // other, self.end // other)

    def __and__(self, other: Self) -> Self:
        return Interval(max(self.start, other.start), min(self.end, other.end))

    def lerp(self, x: float, /) -> float:
        return lerp(self.start, self.end, x)

    def lerp_clamped(self, x: float, /) -> float:
        return lerp_clamped(self.start, self.end, x)

    def unlerp(self, x: float, /) -> float:
        return unlerp(self.start, self.end, x)

    def unlerp_clamped(self, x: float, /) -> float:
        return unlerp_clamped(self.start, self.end, x)


@native_function(Op.Lerp)
def lerp(a, b, x, /):
    return a + (b - a) * x


@native_function(Op.LerpClamped)
def lerp_clamped(a, b, x, /):
    return a + (b - a) * max(0, min(1, x))


@native_function(Op.Unlerp)
def unlerp(a, b, x, /):
    return (x - a) / (b - a)


@native_function(Op.UnlerpClamped)
def unlerp_clamped(a, b, x, /):
    return max(0, min(1, (x - a) / (b - a)))


@native_function(Op.Remap)
def remap(a, b, c, d, x, /):
    return c + (d - c) * (x - a) / (b - a)


@native_function(Op.RemapClamped)
def remap_clamped(a, b, c, d, x, /):
    return c + (d - c) * max(0, min(1, (x - a) / (b - a)))



from sonolus.script.iterator import ArrayLike, SonolusIterator
from sonolus.script.num import Num
from sonolus.script.record import Record


class Range(Record, ArrayLike[Num]):
    start: int
    end: int
    step: int

    def __new__(cls, start: Num, end: Num | None = None, step: Num = 1):
        if end is None:
            start, end = 0, start
        return super().__new__(cls, start, end, step)

    def __iter__(self) -> SonolusIterator:
        return RangeIterator(self.start, self.end, self.step)

    def __contains__(self, item):
        if self.step > 0:
            return self.start <= item < self.end and (item - self.start) % self.step == 0
        else:
            return self.end < item <= self.start and (self.start - item) % -self.step == 0

    def size(self) -> int:
        if self.step > 0:
            diff = self.end - self.start
            if diff <= 0:
                return 0
            return (diff + self.step - 1) // self.step
        else:
            diff = self.start - self.end
            if diff <= 0:
                return 0
            return (diff - self.step - 1) // -self.step

    def __getitem__(self, index: Num) -> Num:
        return self.start + index * self.step

    def __setitem__(self, index: Num, value: Num):
        raise TypeError("Range does not support item assignment")


class RangeIterator(Record, SonolusIterator):
    value: int
    end: int
    step: int

    def has_next(self) -> bool:
        if self.step > 0:
            return self.value < self.end
        else:
            return self.value > self.end

    def next(self) -> Num:
        value = self.value
        self.value += self.step
        return value



# ruff: noqa: A002
from dataclasses import dataclass
from typing import Annotated, Any, NewType, dataclass_transform, get_origin

from sonolus.backend.mode import Mode
from sonolus.backend.place import BlockPlace
from sonolus.script.debug import assert_unreachable
from sonolus.script.internal.context import ctx
from sonolus.script.internal.descriptor import SonolusDescriptor
from sonolus.script.internal.generic import validate_concrete_type
from sonolus.script.internal.introspection import get_field_specifiers
from sonolus.script.num import Num


@dataclass
class SliderOption:
    name: str | None
    standard: bool
    advanced: bool
    scope: str | None
    default: float
    min: float
    max: float
    step: float
    unit: str | None

    def to_dict(self):
        result = {
            "type": "slider",
            "name": self.name,
            "standard": self.standard,
            "advanced": self.advanced,
            "def": self.default,
            "min": self.min,
            "max": self.max,
            "step": self.step,
        }
        if self.scope is not None:
            result["scope"] = self.scope
        if self.unit is not None:
            result["unit"] = self.unit
        return result


@dataclass
class ToggleOption:
    name: str | None
    standard: bool
    advanced: bool
    scope: str | None
    default: bool

    def to_dict(self):
        result = {
            "type": "toggle",
            "name": self.name,
            "standard": self.standard,
            "advanced": self.advanced,
            "def": int(self.default),
        }
        if self.scope is not None:
            result["scope"] = self.scope
        return result


@dataclass
class SelectOption:
    name: str | None
    standard: bool
    advanced: bool
    scope: str | None
    default: str
    values: list[str]

    def to_dict(self):
        result = {
            "type": "select",
            "name": self.name,
            "standard": self.standard,
            "advanced": self.advanced,
            "def": self.values.index(self.default),
            "values": self.values,
        }
        if self.scope is not None:
            result["scope"] = self.scope
        return result


def slider_option(
    *,
    name: str | None = None,
    standard: bool = False,
    advanced: bool = False,
    default: float,
    min: float,
    max: float,
    step: float,
    unit: str | None = None,
    scope: str | None = None,
) -> Any:
    return SliderOption(name, standard, advanced, scope, default, min, max, step, unit)


def toggle_option(
    *,
    name: str | None = None,
    standard: bool = False,
    advanced: bool = False,
    default: bool,
    scope: str | None = None,
) -> Any:
    return ToggleOption(name, standard, advanced, scope, default)


def select_option(
    *,
    name: str | None = None,
    standard: bool = False,
    advanced: bool = False,
    default: str,
    values: list[str],
    scope: str | None = None,
) -> Any:
    return SelectOption(name, standard, advanced, scope, default, values)


type Options = NewType("Options", Any)
type OptionInfo = SliderOption | ToggleOption | SelectOption


class OptionField(SonolusDescriptor):
    info: OptionInfo
    index: int

    def __init__(self, info: OptionInfo, index: int):
        self.info = info
        self.index = index

    def __get__(self, instance, owner):
        if ctx():
            match ctx().global_state.mode:
                case Mode.Play:
                    block = ctx().blocks.LevelOption
                case Mode.Watch:
                    block = ctx().blocks.LevelOption
                case Mode.Preview:
                    block = ctx().blocks.PreviewOption
                case Mode.Tutorial:
                    block = None
                case _:
                    assert_unreachable()
            if block is not None:
                return Num._from_place_(BlockPlace(block, self.index))
            else:
                return Num._accept_(self.info.default)

    def __set__(self, instance, value):
        raise AttributeError("Options are read-only")


@dataclass_transform()
def options[T](cls: type[T]) -> T | Options:
    if len(cls.__bases__) != 1:
        raise ValueError("Options class must not inherit from any class (except object)")
    instance = cls()
    entries = []
    for i, (name, annotation) in enumerate(get_field_specifiers(cls).items()):
        if get_origin(annotation) is not Annotated:
            raise TypeError(f"Invalid annotation for options: {annotation}")
        annotation_type = annotation.__args__[0]
        annotation_values = annotation.__metadata__
        if len(annotation_values) != 1:
            raise ValueError("Invalid annotation values for options")
        annotation_type = validate_concrete_type(annotation_type)
        if annotation_type is not Num:
            raise TypeError(f"Invalid annotation type for options: {annotation_type}")
        annotation_value = annotation_values[0]
        if not isinstance(annotation_value, SliderOption | ToggleOption | SelectOption):
            raise TypeError(f"Invalid annotation value for options: {annotation_value}")
        if annotation_value.name is None:
            annotation_value.name = name
        entries.append(annotation_value)
        setattr(cls, name, OptionField(annotation_value, i))
    instance._options_ = entries
    instance._is_comptime_value_ = True
    return instance


@options
class EmptyOptions:
    pass



import gzip
import json
import struct
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path

from sonolus.backend.mode import Mode
from sonolus.build.compile import compile_mode
from sonolus.build.defaults import EMPTY_ENGINE_PREVIEW_DATA, EMPTY_ENGINE_TUTORIAL_DATA
from sonolus.script.archetype import BaseArchetype
from sonolus.script.bucket import Buckets
from sonolus.script.callbacks import update_spawn_callback
from sonolus.script.effect import Effects
from sonolus.script.engine import EngineData
from sonolus.script.internal.context import ReadOnlyMemory
from sonolus.script.options import Options
from sonolus.script.particle import Particles
from sonolus.script.sprite import Skin
from sonolus.script.ui import UiConfig

type JsonValue = None | bool | int | float | str | list[JsonValue] | dict[str, JsonValue]


@dataclass
class PackagedEngine:
    configuration: bytes
    play_data: bytes
    watch_data: bytes
    preview_data: bytes
    tutorial_data: bytes
    rom: bytes

    def write(self, path: Path):
        path.mkdir(parents=True, exist_ok=True)
        (path / "EngineConfiguration").write_bytes(self.configuration)
        (path / "EnginePlayData").write_bytes(self.play_data)
        (path / "EngineWatchData").write_bytes(self.watch_data)
        (path / "EnginePreviewData").write_bytes(self.preview_data)
        (path / "EngineTutorialData").write_bytes(self.tutorial_data)
        (path / "EngineRom").write_bytes(self.rom)


def package_engine(engine: EngineData):
    rom = ReadOnlyMemory()
    configuration = build_engine_configuration(engine.options, engine.ui)
    play_data = build_play_mode(
        archetypes=engine.play.archetypes,
        skin=engine.play.skin,
        effects=engine.play.effects,
        particles=engine.play.particles,
        buckets=engine.play.buckets,
        rom=rom,
    )
    watch_data = build_watch_mode(
        archetypes=engine.watch.archetypes,
        skin=engine.watch.skin,
        effects=engine.watch.effects,
        particles=engine.watch.particles,
        buckets=engine.watch.buckets,
        rom=rom,
        update_spawn=engine.watch.update_spawn,
    )
    return PackagedEngine(
        configuration=package_output(configuration),
        play_data=package_output(play_data),
        watch_data=package_output(watch_data),
        preview_data=package_output(EMPTY_ENGINE_PREVIEW_DATA),
        tutorial_data=package_output(EMPTY_ENGINE_TUTORIAL_DATA),
        rom=package_rom(rom),
    )


def build_engine_configuration(
    options: Options,
    ui: UiConfig,
) -> JsonValue:
    return {
        "options": [option.to_dict() for option in options._options_],
        "ui": ui.to_dict(),
    }


def build_play_mode(
    archetypes: list[type[BaseArchetype]],
    skin: Skin,
    effects: Effects,
    particles: Particles,
    buckets: Buckets,
    rom: ReadOnlyMemory,
):
    return {
        **compile_mode(mode=Mode.Play, rom=rom, archetypes=archetypes, global_callbacks=None),
        "skin": build_skin(skin),
        "effect": build_effects(effects),
        "particle": build_particles(particles),
        "buckets": build_buckets(buckets),
    }


def build_watch_mode(
    archetypes: list[type[BaseArchetype]],
    skin: Skin,
    effects: Effects,
    particles: Particles,
    buckets: Buckets,
    rom: ReadOnlyMemory,
    update_spawn: Callable[[], float],
):
    return {
        **compile_mode(
            mode=Mode.Watch, rom=rom, archetypes=archetypes, global_callbacks=[(update_spawn_callback, update_spawn)]
        ),
        "skin": build_skin(skin),
        "effect": build_effects(effects),
        "particle": build_particles(particles),
        "buckets": build_buckets(buckets),
    }


def build_skin(skin: Skin) -> JsonValue:
    return {"sprites": [{"name": name, "id": i} for i, name in enumerate(skin._sprites_)]}


def build_effects(effects: Effects) -> JsonValue:
    return {"clips": [{"name": name, "id": i} for i, name in enumerate(effects._effects_)]}


def build_particles(particles: Particles) -> JsonValue:
    return {"effects": [{"name": name, "id": i} for i, name in enumerate(particles._particles_)]}


def build_buckets(buckets: Buckets) -> JsonValue:
    return [bucket.to_dict() for bucket in buckets._buckets_]


def package_rom(rom: ReadOnlyMemory) -> bytes:
    values = rom.values or [0]
    output = bytearray()

    for value in values:
        output.extend(struct.pack("<f", value))

    return gzip.compress(bytes(output))


def package_output(value: JsonValue) -> bytes:
    json_data = json.dumps(value, separators=(",", ":")).encode("utf-8")
    return gzip.compress(json_data)



from collections.abc import Callable
from typing import Any, Never

from sonolus.backend.flow import cfg_to_mermaid
from sonolus.backend.mode import Mode
from sonolus.backend.ops import Op
from sonolus.backend.simplify import CoalesceFlow
from sonolus.script.comptime import Comptime
from sonolus.script.internal.context import GlobalContextState, ctx, set_ctx
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.native import native_function
from sonolus.script.num import Num
from sonolus.script.values import with_default


@meta_fn
def error(message: str | None = None) -> None:
    message = Comptime._accept_(message)._as_py_() if message is not None else "Error"
    if not isinstance(message, str):
        raise ValueError("Expected a string")
    if ctx():
        debug_log(ctx().map_constant(message))
        debug_pause()
        terminate()
    else:
        raise RuntimeError(message)


@native_function(Op.DebugLog)
def debug_log(value: Num):
    print(f"[DEBUG] {value}")


@native_function(Op.DebugPause)
def debug_pause():
    input("[DEBUG] Paused")


def assert_true(value: Num, message: str | None = None):
    message = with_default(message, "Assertion failed")
    if not value:
        error(message)


def assert_false(value: Num, message: str | None = None):
    message = with_default(message, "Assertion failed")
    if value:
        error(message)


@meta_fn
def assert_unreachable(message: str | None = None) -> Never:
    message = validate_value(message)._as_py_() or "Unreachable code reached"
    raise RuntimeError(message)


@meta_fn
def terminate():
    if ctx():
        set_ctx(ctx().into_dead())
    else:
        raise RuntimeError("Terminated")


def visualize_cfg(fn: Callable[[], Any]) -> str:
    from sonolus.build.compile import callback_to_cfg

    cfg = callback_to_cfg(GlobalContextState(Mode.Play), fn, "")
    cfg = CoalesceFlow().run(cfg)
    return cfg_to_mermaid(cfg)



from sonolus.backend.ops import Op
from sonolus.script.internal.native import native_function


@native_function(Op.BeatToBPM)
def beat_to_bpm(beat: float) -> float:
    raise NotImplementedError


@native_function(Op.BeatToTime)
def beat_to_time(beat: float) -> float:
    raise NotImplementedError


@native_function(Op.BeatToStartingBeat)
def beat_to_starting_beat(beat: float) -> float:
    raise NotImplementedError


@native_function(Op.BeatToStartingTime)
def beat_to_starting_time(beat: float) -> float:
    raise NotImplementedError


@native_function(Op.TimeToScaledTime)
def time_to_scaled_time(time: float) -> float:
    raise NotImplementedError


@native_function(Op.TimeToStartingScaledTime)
def time_to_starting_scaled_time(time: float) -> float:
    raise NotImplementedError


@native_function(Op.TimeToStartingTime)
def time_to_starting_time(time: float) -> float:
    raise NotImplementedError


@native_function(Op.TimeToTimeScale)
def time_to_timescale(time: float) -> float:
    raise NotImplementedError



import math

from sonolus.backend.ops import Op
from sonolus.script.internal.native import native_function


@native_function(Op.Sin)
def sin(x: float) -> float:
    return math.sin(x)


@native_function(Op.Cos)
def cos(x: float) -> float:
    return math.cos(x)


@native_function(Op.Tan)
def tan(x: float) -> float:
    return math.tan(x)


@native_function(Op.Arcsin)
def asin(x: float) -> float:
    return math.asin(x)


@native_function(Op.Arccos)
def acos(x: float) -> float:
    return math.acos(x)


@native_function(Op.Arctan)
def atan(x: float) -> float:
    return math.atan(x)


@native_function(Op.Arctan2)
def atan2(y: float, x: float) -> float:
    return math.atan2(y, x)


@native_function(Op.Sinh)
def sinh(x: float) -> float:
    return math.sinh(x)


@native_function(Op.Cosh)
def cosh(x: float) -> float:
    return math.cosh(x)


@native_function(Op.Tanh)
def tanh(x: float) -> float:
    return math.tanh(x)


@native_function(Op.Floor)
def floor(x: float) -> float:
    return math.floor(x)


@native_function(Op.Ceil)
def ceil(x: float) -> float:
    return math.ceil(x)


@native_function(Op.Trunc)
def trunc(x: float) -> float:
    return math.trunc(x)


@native_function(Op.Round)
def _round(x: float) -> float:
    return round(x)


MATH_BUILTIN_IMPLS = {
    id(math.sin): sin,
    id(math.cos): cos,
    id(math.tan): tan,
    id(math.asin): asin,
    id(math.acos): acos,
    id(math.atan): atan,
    id(math.atan2): atan2,
    id(math.sinh): sinh,
    id(math.cosh): cosh,
    id(math.tanh): tanh,
    id(math.floor): floor,
    id(math.ceil): ceil,
    id(math.trunc): trunc,
    id(round): _round,
}



from sonolus.script.internal.context import ctx
from sonolus.script.internal.generic import validate_concrete_type
from sonolus.script.internal.impl import meta_fn, validate_value


@meta_fn
def alloc[T](type_: type[T]) -> T:
    """Returns an uninitialized instance of the given type."""
    type_ = validate_concrete_type(type_)
    if ctx():
        return type_._alloc_()
    else:
        return type_._alloc_()._as_py_()


@meta_fn
def zeros[T](type_: type[T]) -> T:
    """Returns a new instance of the given type initialized with zeros."""
    type_ = validate_concrete_type(type_)
    if ctx():
        return copy(type_._from_list_([0] * type_._size_()))
    else:
        return type_._from_list_([0] * type_._size_())._as_py_()


@meta_fn
def copy[T](value: T) -> T:
    """Returns a deep copy of the given value."""
    value = validate_value(value)
    if ctx():
        return value._copy_()
    else:
        return value._copy_()._as_py_()


@meta_fn
def with_default[T](value: T, default: T) -> T:
    """Returns the given value if it's not None, otherwise the default value."""
    value = validate_value(value)
    default = validate_value(default)
    if value._is_py_() and value._as_py_() is None:
        return default
    return value



from __future__ import annotations

from collections.abc import Callable

from sonolus.build.collection import Asset
from sonolus.script.archetype import BaseArchetype
from sonolus.script.bucket import Buckets, EmptyBuckets
from sonolus.script.effect import Effects, EmptyEffects
from sonolus.script.options import EmptyOptions, Options
from sonolus.script.particle import EmptyParticles, Particles
from sonolus.script.sprite import EmptySkin, Skin
from sonolus.script.ui import UiConfig


class Engine:
    version = 12

    def __init__(
        self,
        *,
        name: str,
        title: str | None = None,
        subtitle: str = "Sonolus.py Engine",
        author: str = "Unknown",
        skin: str | None = None,
        background: str | None = None,
        effect: str | None = None,
        particle: str | None = None,
        thumbnail: Asset | None = None,
        data: EngineData,
    ) -> None:
        self.name = name
        self.title = title or name
        self.subtitle = subtitle
        self.author = author
        self.skin = skin
        self.background = background
        self.effect = effect
        self.particle = particle
        self.thumbnail = thumbnail
        self.data = data


def default_callback() -> float:
    return 0.0


class PlayMode:
    def __init__(
        self,
        *,
        archetypes: list[type[BaseArchetype]] | None = None,
        skin: Skin = EmptySkin,
        effects: Effects = EmptyEffects,
        particles: Particles = EmptyParticles,
        buckets: Buckets = EmptyBuckets,
    ) -> None:
        self.archetypes = archetypes or []
        self.skin = skin
        self.effects = effects
        self.particles = particles
        self.buckets = buckets


class WatchMode:
    def __init__(
        self,
        *,
        archetypes: list[type[BaseArchetype]] | None = None,
        skin: Skin = EmptySkin,
        effects: Effects = EmptyEffects,
        particles: Particles = EmptyParticles,
        buckets: Buckets = EmptyBuckets,
        update_spawn: Callable[[], float],
    ) -> None:
        self.archetypes = archetypes or []
        self.skin = skin
        self.effects = effects
        self.particles = particles
        self.buckets = buckets
        self.update_spawn = update_spawn


class EngineData:
    ui: UiConfig
    options: Options
    play: PlayMode
    watch: WatchMode

    def __init__(
        self,
        *,
        ui: UiConfig | None = None,
        options: Options = EmptyOptions,
        play: PlayMode | None = None,
        watch: WatchMode | None = None,
    ) -> None:
        self.ui = ui or UiConfig()
        self.options = options
        self.play = play or PlayMode()
        self.watch = watch or WatchMode(update_spawn=default_callback)



from typing import Self

from sonolus.script.math import atan2
from sonolus.script.num import Num
from sonolus.script.record import Record


class Vec2(Record):
    x: float
    y: float

    @property
    def magnitude(self) -> Num:
        return (self.x**2 + self.y**2) ** 0.5

    @property
    def angle(self) -> Num:
        return atan2(self.y, self.x)

    def dot(self, other: Self) -> Num:
        return self.x * other.x + self.y * other.y

    @property
    def tuple(self) -> tuple[float, float]:
        return self.x, self.y

    def __add__(self, other: Self) -> Self:
        return Vec2(x=self.x + other.x, y=self.y + other.y)

    def __sub__(self, other: Self) -> Self:
        return Vec2(x=self.x - other.x, y=self.y - other.y)

    def __mul__(self, other: Self | float) -> Self:
        match other:
            case Vec2(x, y):
                return Vec2(x=self.x * x, y=self.y * y)
            case float() | int() as factor:
                return Vec2(x=self.x * factor, y=self.y * factor)

    def __truediv__(self, other: Self | float) -> Self:
        match other:
            case Vec2(x, y):
                return Vec2(x=self.x / x, y=self.y / y)
            case float() | int() as factor:
                return Vec2(x=self.x / factor, y=self.y / factor)

    def __neg__(self) -> Self:
        return Vec2(x=-self.x, y=-self.y)



from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from typing import Any, Self

from sonolus.backend.blocks import BlockData, PlayBlock
from sonolus.backend.flow import BasicBlock, FlowEdge
from sonolus.backend.ir import IRConst, IRStmt
from sonolus.backend.mode import Mode
from sonolus.backend.place import Block, BlockPlace, TempBlock
from sonolus.script.globals import GlobalInfo, GlobalPlaceholder
from sonolus.script.internal.value import Value

_compiler_internal_ = True

context_var = ContextVar("context_var", default=None)


class GlobalContextState:
    archetypes: dict[type, int]
    rom: ReadOnlyMemory
    const_mappings: dict[Any, int]
    environment_mappings: dict[GlobalInfo, int]
    environment_offsets: dict[Block, int]
    mode: Mode

    def __init__(self, mode: Mode, archetypes: dict[type, int] | None = None, rom: ReadOnlyMemory | None = None):
        self.archetypes = archetypes or {}
        self.rom = ReadOnlyMemory() if rom is None else rom
        self.const_mappings = {}
        self.environment_mappings = {}
        self.environment_offsets = {}
        self.mode = mode


class CallbackContextState:
    callback: str
    used_names: dict[str, int]

    def __init__(self, callback: str):
        self.callback = callback
        self.used_names = {}


class Context:
    global_state: GlobalContextState
    callback_state: CallbackContextState
    statements: list[IRStmt]
    test: IRStmt
    outgoing: dict[float | None, Context]
    scope: Scope
    loop_variables: dict[str, Value]
    live: bool

    def __init__(
        self,
        global_state: GlobalContextState,
        callback_state: CallbackContextState,
        scope: Scope | None = None,
        live: bool = True,
    ):
        self.global_state = global_state
        self.callback_state = callback_state
        self.statements = []
        self.test = IRConst(0)
        self.outgoing = {}
        self.scope = scope if scope is not None else Scope()
        self.loop_variables = {}
        self.live = live

    @property
    def rom(self) -> ReadOnlyMemory:
        return self.global_state.rom

    @property
    def blocks(self) -> type[Block]:
        return self.global_state.mode.blocks

    @property
    def callback(self) -> str:
        return self.callback_state.callback

    @property
    def used_names(self) -> dict[str, int]:
        return self.callback_state.used_names

    @property
    def const_mappings(self) -> dict[Any, int]:
        return self.global_state.const_mappings

    def check_readable(self, place: BlockPlace):
        if not self.callback:
            return
        if isinstance(place.block, BlockData) and self.callback not in self.blocks(place.block).readable:
            raise RuntimeError(f"Block {place.block} is not readable in {self.callback}")

    def check_writable(self, place: BlockPlace):
        if not self.callback:
            return
        if isinstance(place.block, BlockData) and self.callback not in self.blocks(place.block).writable:
            raise RuntimeError(f"Block {place.block} is not writable in {self.callback}")

    def add_statements(self, *statements: IRStmt):
        self.statements.extend(statements)

    def alloc(self, name: str | None = None, size: int = 1) -> BlockPlace:
        if size == 0:
            return BlockPlace(TempBlock(name or "e", 0), 0)
        name = name or ("v" if size == 1 else "a")
        num = self._get_alloc_number(name)
        return BlockPlace(TempBlock(f"{name}{num}", size), 0)

    def _get_alloc_number(self, name: str) -> int:
        if name not in self.used_names:
            self.used_names[name] = 0
        self.used_names[name] += 1
        return self.used_names[name]

    def copy_with_scope(self, scope: Scope) -> Context:
        return Context(
            global_state=self.global_state,
            callback_state=self.callback_state,
            scope=scope,
            live=self.live,
        )

    def branch(self, condition: float | None):
        assert condition not in self.outgoing
        result = self.copy_with_scope(self.scope.copy())
        self.outgoing[condition] = result
        return result

    def branch_with_scope(self, condition: float | None, scope: Scope):
        assert condition not in self.outgoing
        result = self.copy_with_scope(scope)
        self.outgoing[condition] = result
        return result

    def into_dead(self):
        """Create a new context for code that is unreachable, like after a return statement."""
        result = self.copy_with_scope(self.scope.copy())
        result.live = False
        return result

    def prepare_loop_header(self, to_merge: set[str]) -> Self:
        # to_merge is the set of bindings set anywhere in the loop
        # we need to invalidate them in the header if they're reference types
        # or merge them if they're value types
        # structure is self -> intermediate -> header -> body (continue -> header) | exit
        assert len(self.outgoing) == 0
        header = self.branch(None)
        for name in to_merge:
            binding = self.scope.get_binding(name)
            if not isinstance(binding, ValueBinding):
                continue
            value = binding.value
            type_ = type(value)
            if type_._is_value_type_():
                target_value = type_._from_place_(header.alloc(size=type_._size_()))
                with using_ctx(self):
                    target_value._set_(value)
                header.scope.set_value(name, target_value)
                header.loop_variables[name] = target_value
            else:
                header.scope.set_binding(name, ConflictBinding())
        return header

    def branch_to_loop_header(self, header: Self):
        if not self.live:
            return
        assert len(self.outgoing) == 0
        self.outgoing[None] = header
        for name, target_value in header.loop_variables.items():
            with using_ctx(self):
                value = self.scope.get_value(name)
                value = type(target_value)._accept_(value)
                target_value._set_(value)

    def map_constant(self, value: Any) -> int:
        if value not in self.const_mappings:
            self.const_mappings[value] = len(self.const_mappings)
        return self.const_mappings[value]

    def get_global_base(self, value: GlobalInfo | GlobalPlaceholder) -> BlockPlace:
        block = value.blocks.get(self.global_state.mode)
        if block is None:
            raise RuntimeError(f"Global {value.name} is not available in '{self.global_state.mode.name}' mode")
        if value not in self.global_state.environment_mappings:
            if value.offset is None:
                offset = self.global_state.environment_offsets.get(block, 0)
                self.global_state.environment_mappings[value] = offset
                self.global_state.environment_offsets[block] = offset + value.size
            else:
                self.global_state.environment_mappings[value] = value.offset
        return BlockPlace(block, self.global_state.environment_mappings[value])

    @classmethod
    def meet(cls, contexts: list[Context]) -> Context:
        if not contexts:
            raise RuntimeError("Cannot meet empty list of contexts")
        if not any(context.live for context in contexts):
            return contexts[0].into_dead()
        contexts = [context for context in contexts if context.live]
        assert not any(context.outgoing for context in contexts)
        assert all(len(context.outgoing) == 0 for context in contexts)
        target = contexts[0].copy_with_scope(Scope())
        Scope.apply_merge(target, contexts)
        for context in contexts:
            context.outgoing[None] = target
        return target


def ctx() -> Context | None:
    return context_var.get()


def set_ctx(value: Context | None):
    return context_var.set(value)


@contextmanager
def using_ctx(value: Context | None):
    token = context_var.set(value)
    try:
        yield
    finally:
        context_var.reset(token)


class ReadOnlyMemory:
    values: list[float]
    indexes: dict[tuple[float, ...], int]

    def __init__(self):
        self.values = []
        self.indexes = {}

    def __getitem__(self, item: tuple[float, ...]) -> BlockPlace:
        if item not in self.indexes:
            index = len(self.values)
            self.indexes[item] = index
            self.values.extend(item)
        else:
            index = self.indexes[item]
        return BlockPlace(self.block, index)

    @property
    def block(self) -> Block:
        if ctx():
            return ctx().blocks.EngineRom
        else:
            return PlayBlock.EngineRom


@dataclass
class ValueBinding:
    value: Value


@dataclass
class ConflictBinding:
    pass


@dataclass
class EmptyBinding:
    pass


Binding = ValueBinding | ConflictBinding | EmptyBinding


class Scope:
    bindings: dict[str, Binding]

    def __init__(self, bindings: dict[str, Binding] | None = None):
        self.bindings = bindings or {}

    def get_binding(self, name: str) -> Binding:
        return self.bindings.get(name, EmptyBinding())

    def set_binding(self, name: str, binding: Binding):
        self.bindings[name] = binding

    def get_value(self, name: str) -> Value:
        binding = self.get_binding(name)
        match binding:
            case ValueBinding(value):
                # we don't need to call _get_() here because _set_() is never called where it could be a problem
                return value
            case ConflictBinding():
                raise RuntimeError(
                    f"Binding '{name}' has multiple conflicting definitions or may not be guaranteed to be defined"
                )
            case EmptyBinding():
                raise RuntimeError(f"Binding '{name}' is not defined")

    def set_value(self, name: str, value: Value):
        from sonolus.script.internal.impl import validate_value

        self.bindings[name] = ValueBinding(validate_value(value))

    def delete_binding(self, name: str):
        del self.bindings[name]

    def copy(self) -> Scope:
        return Scope(self.bindings.copy())

    @classmethod
    def apply_merge(cls, target: Context, incoming: list[Context]):
        if not incoming:
            return
        assert all(len(inc.outgoing) == 0 for inc in incoming)
        sources = [context.scope for context in incoming]
        keys = {key for source in sources for key in source.bindings}
        for key in keys:
            bindings = [source.get_binding(key) for source in sources]
            if not all(isinstance(binding, ValueBinding) for binding in bindings):
                target.scope.set_binding(key, ConflictBinding())
                continue
            values = [binding.value for binding in bindings]
            if len({id(value) for value in values}) == 1:
                target.scope.set_binding(key, ValueBinding(values[0]))
                continue
            types = {type(value) for value in values}
            if len(types) > 1:
                target.scope.set_binding(key, ConflictBinding())
                continue
            common_type: type[Value] = types.pop()
            if common_type._is_value_type_():
                target_value = common_type._from_place_(target.alloc(size=common_type._size_()))
                for inc in incoming:
                    with using_ctx(inc):
                        target_value._set_(inc.scope.get_value(key))
                target.scope.set_value(key, target_value)
                continue
            else:
                target.scope.set_binding(key, ConflictBinding())
                continue


def iter_contexts(context: Context):
    seen = set()
    queue = [context]
    while queue:
        current = queue.pop()
        if current in seen:
            continue
        seen.add(current)
        yield current
        queue.extend(current.outgoing.values())


def context_to_cfg(context: Context) -> BasicBlock:
    blocks = {context: BasicBlock(statements=context.statements, test=context.test)}
    for current in iter_contexts(context):
        for condition, target in current.outgoing.items():
            if target not in blocks:
                blocks[target] = BasicBlock(statements=target.statements, test=target.test)
            edge = FlowEdge(src=blocks[current], dst=blocks[target], cond=condition)
            blocks[current].outgoing.add(edge)
            blocks[target].incoming.add(edge)
    return blocks[context]



from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated, Any, NewType, dataclass_transform, get_origin

from sonolus.backend.ops import Op
from sonolus.script.graphics import QuadLike, flatten_quad
from sonolus.script.internal.introspection import get_field_specifiers
from sonolus.script.internal.native import native_function
from sonolus.script.record import Record


class Particle(Record):
    id: int

    def is_available(self) -> bool:
        return _has_particle_effect(self.id)

    def spawn(self, quad: QuadLike, duration: float, loop: bool = False) -> ParticleHandle:
        return ParticleHandle(_spawn_particle_effect(self.id, *flatten_quad(quad), duration, loop))


class ParticleHandle(Record):
    id: int

    def move(self, quad: QuadLike) -> None:
        _move_particle_effect(self.id, *flatten_quad(quad))

    def destroy(self) -> None:
        _destroy_particle_effect(self.id)


@native_function(Op.HasParticleEffect)
def _has_particle_effect(particle_id: int) -> bool:
    raise NotImplementedError


@native_function(Op.SpawnParticleEffect)
def _spawn_particle_effect(
    particle_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    duration: float,
    loop: bool,
) -> int:
    raise NotImplementedError


@native_function(Op.MoveParticleEffect)
def _move_particle_effect(
    handle: int, x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, x4: float, y4: float
) -> None:
    raise NotImplementedError


@native_function(Op.DestroyParticleEffect)
def _destroy_particle_effect(handle: int) -> None:
    raise NotImplementedError


@dataclass
class ParticleInfo:
    name: str


def particle(name: str) -> Any:
    return ParticleInfo(name)


type Particles = NewType("Particles", Any)


@dataclass_transform()
def particles[T](cls: type[T]) -> T | Particles:
    if len(cls.__bases__) != 1:
        raise ValueError("Particles class must not inherit from any class (except object)")
    instance = cls()
    names = []
    for i, (name, annotation) in enumerate(get_field_specifiers(cls).items()):
        if get_origin(annotation) is not Annotated:
            raise TypeError(f"Invalid annotation for particles: {annotation}")
        annotation_type = annotation.__args__[0]
        annotation_values = annotation.__metadata__
        if annotation_type is not Particle:
            raise TypeError(f"Invalid annotation for particles: {annotation}, expected annotation of type Particle")
        if len(annotation_values) != 1 or not isinstance(annotation_values[0], ParticleInfo):
            raise TypeError(
                f"Invalid annotation for particles: {annotation}, expected a single string annotation value"
            )
        particle_name = annotation_values[0].name
        names.append(particle_name)
        setattr(instance, name, Particle(i))
    instance._particles_ = names
    instance._is_comptime_value_ = True
    return instance


class StandardParticle:
    NoteCircularTapNeutral = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_NEUTRAL")]
    NoteCircularTapRed = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_RED")]
    NoteCircularTapGreen = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_GREEN")]
    NoteCircularTapBlue = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_BLUE")]
    NoteCircularTapYellow = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_YELLOW")]
    NoteCircularTapPurple = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_PURPLE")]
    NoteCircularTapCyan = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_CYAN")]
    NoteCircularAlternativeNeutral = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_NEUTRAL")]
    NoteCircularAlternativeRed = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_RED")]
    NoteCircularAlternativeGreen = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_GREEN")]
    NoteCircularAlternativeBlue = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_BLUE")]
    NoteCircularAlternativeYellow = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_YELLOW")]
    NoteCircularAlternativePurple = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_PURPLE")]
    NoteCircularAlternativeCyan = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_CYAN")]
    NoteCircularHoldNeutral = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_NEUTRAL")]
    NoteCircularHoldRed = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_RED")]
    NoteCircularHoldGreen = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_GREEN")]
    NoteCircularHoldBlue = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_BLUE")]
    NoteCircularHoldYellow = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_YELLOW")]
    NoteCircularHoldPurple = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_PURPLE")]
    NoteCircularHoldCyan = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_CYAN")]
    NoteLinearTapNeutral = Annotated[Particle, particle("#NOTE_LINEAR_TAP_NEUTRAL")]
    NoteLinearTapRed = Annotated[Particle, particle("#NOTE_LINEAR_TAP_RED")]
    NoteLinearTapGreen = Annotated[Particle, particle("#NOTE_LINEAR_TAP_GREEN")]
    NoteLinearTapBlue = Annotated[Particle, particle("#NOTE_LINEAR_TAP_BLUE")]
    NoteLinearTapYellow = Annotated[Particle, particle("#NOTE_LINEAR_TAP_YELLOW")]
    NoteLinearTapPurple = Annotated[Particle, particle("#NOTE_LINEAR_TAP_PURPLE")]
    NoteLinearTapCyan = Annotated[Particle, particle("#NOTE_LINEAR_TAP_CYAN")]
    NoteLinearAlternativeNeutral = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_NEUTRAL")]
    NoteLinearAlternativeRed = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_RED")]
    NoteLinearAlternativeGreen = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_GREEN")]
    NoteLinearAlternativeBlue = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_BLUE")]
    NoteLinearAlternativeYellow = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_YELLOW")]
    NoteLinearAlternativePurple = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_PURPLE")]
    NoteLinearAlternativeCyan = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_CYAN")]
    NoteLinearHoldNeutral = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_NEUTRAL")]
    NoteLinearHoldRed = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_RED")]
    NoteLinearHoldGreen = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_GREEN")]
    NoteLinearHoldBlue = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_BLUE")]
    NoteLinearHoldYellow = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_YELLOW")]
    NoteLinearHoldPurple = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_PURPLE")]
    NoteLinearHoldCyan = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_CYAN")]
    LaneCircular = Annotated[Particle, particle("#LANE_CIRCULAR")]
    LaneLinear = Annotated[Particle, particle("#LANE_LINEAR")]
    SlotCircular = Annotated[Particle, particle("#SLOT_CIRCULAR")]
    SlotLinear = Annotated[Particle, particle("#SLOT_LINEAR")]
    JudgeLineCircular = Annotated[Particle, particle("#JUDGE_LINE_CIRCULAR")]
    JudgeLineLinear = Annotated[Particle, particle("#JUDGE_LINE_LINEAR")]


@particles
class EmptyParticles:
    pass



EMPTY_ENGINE_PLAY_DATA = {
    "skin": {"sprites": []},
    "effect": {"clips": []},
    "particle": {"effects": []},
    "buckets": [],
    "archetypes": [],
    "nodes": [{"value": 0}],
}

EMPTY_ENGINE_WATCH_DATA = {
    "skin": {"sprites": []},
    "effect": {"clips": []},
    "particle": {"effects": []},
    "buckets": [],
    "archetypes": [],
    "updateSpawn": 0,
    "nodes": [{"value": 0}],
}

EMPTY_ENGINE_PREVIEW_DATA = {
    "skin": {"sprites": []},
    "archetypes": [],
    "nodes": [{"value": 0}],
}

EMPTY_ENGINE_TUTORIAL_DATA = {
    "skin": {"sprites": []},
    "effect": {"clips": []},
    "particle": {"effects": []},
    "instruction": {"texts": [], "icons": []},
    "nodes": [{"value": 0}],
}



from sonolus.backend.allocate import AllocateBasic
from sonolus.backend.flow import BasicBlock
from sonolus.backend.simplify import CoalesceFlow


def optimize_and_allocate(cfg: BasicBlock):
    cfg = CoalesceFlow().run(cfg)
    cfg = AllocateBasic().run(cfg)
    return cfg



class StandardText:
    Post = "#POST"
    Playlist = "#PLAYLIST"
    Level = "#LEVEL"
    Skin = "#SKIN"
    Background = "#BACKGROUND"
    Effect = "#EFFECT"
    Particle = "#PARTICLE"
    Engine = "#ENGINE"
    Replay = "#REPLAY"
    Room = "#ROOM"
    Grade = "#GRADE"
    ArcadeScore = "#ARCADE_SCORE"
    AccuracyScore = "#ACCURACY_SCORE"
    Combo = "#COMBO"
    Perfect = "#PERFECT"
    Great = "#GREAT"
    Good = "#GOOD"
    Miss = "#MISS"
    Keywords = "#KEYWORDS"
    Name = "#NAME"
    Rating = "#RATING"
    RatingMinimum = "#RATING_MINIMUM"
    RatingMaximum = "#RATING_MAXIMUM"
    Title = "#TITLE"
    Subtitle = "#SUBTITLE"
    Artists = "#ARTISTS"
    Time = "#TIME"
    Author = "#AUTHOR"
    Genre = "#GENRE"
    Type = "#TYPE"
    Category = "#CATEGORY"
    Language = "#LANGUAGE"
    Difficulty = "#DIFFICULTY"
    Version = "#VERSION"
    Length = "#LENGTH"
    LengthMinimum = "#LENGTH_MINIMUM"
    LengthMaximum = "#LENGTH_MAXIMUM"
    AdditionalInformation = "#ADDITIONAL_INFORMATION"
    Timezone = "#TIMEZONE"
    Region = "#REGION"
    Content = "#CONTENT"
    Comment = "#COMMENT"
    Review = "#REVIEW"
    Reply = "#REPLY"
    Message = "#MESSAGE"
    Role = "#ROLE"
    Permission = "#PERMISSION"
    User = "#USER"
    Speed = "#SPEED"
    Mirror = "#MIRROR"
    Random = "#RANDOM"
    Hidden = "#HIDDEN"
    JudgmentStrict = "#JUDGMENT_STRICT"
    JudgmentLoose = "#JUDGMENT_LOOSE"
    EffectAuto = "#EFFECT_AUTO"
    Stage = "#STAGE"
    StagePosition = "#STAGE_POSITION"
    StageSize = "#STAGE_SIZE"
    StageRotation = "#STAGE_ROTATION"
    StageDirection = "#STAGE_DIRECTION"
    StageAlpha = "#STAGE_ALPHA"
    StageAnimation = "#STAGE_ANIMATION"
    StageTilt = "#STAGE_TILT"
    StageCoverVertical = "#STAGE_COVER_VERTICAL"
    StageCoverHorizontal = "#STAGE_COVER_HORIZONTAL"
    StageCoverAlpha = "#STAGE_COVER_ALPHA"
    StageAspectratioLock = "#STAGE_ASPECTRATIO_LOCK"
    StageEffect = "#STAGE_EFFECT"
    StageEffectPosition = "#STAGE_EFFECT_POSITION"
    StageEffectSize = "#STAGE_EFFECT_SIZE"
    StageEffectRotation = "#STAGE_EFFECT_ROTATION"
    StageEffectDirection = "#STAGE_EFFECT_DIRECTION"
    StageEffectAlpha = "#STAGE_EFFECT_ALPHA"
    Lane = "#LANE"
    LanePosition = "#LANE_POSITION"
    LaneSize = "#LANE_SIZE"
    LaneRotation = "#LANE_ROTATION"
    LaneDirection = "#LANE_DIRECTION"
    LaneAlpha = "#LANE_ALPHA"
    LaneAnimation = "#LANE_ANIMATION"
    LaneEffect = "#LANE_EFFECT"
    LaneEffectPosition = "#LANE_EFFECT_POSITION"
    LaneEffectSize = "#LANE_EFFECT_SIZE"
    LaneEffectRotation = "#LANE_EFFECT_ROTATION"
    LaneEffectDirection = "#LANE_EFFECT_DIRECTION"
    LaneEffectAlpha = "#LANE_EFFECT_ALPHA"
    Judgeline = "#JUDGELINE"
    JudgelinePosition = "#JUDGELINE_POSITION"
    JudgelineSize = "#JUDGELINE_SIZE"
    JudgelineRotation = "#JUDGELINE_ROTATION"
    JudgelineDirection = "#JUDGELINE_DIRECTION"
    JudgelineAlpha = "#JUDGELINE_ALPHA"
    JudgelineAnimation = "#JUDGELINE_ANIMATION"
    JudgelineEffect = "#JUDGELINE_EFFECT"
    JudgelineEffectPosition = "#JUDGELINE_EFFECT_POSITION"
    JudgelineEffectSize = "#JUDGELINE_EFFECT_SIZE"
    JudgelineEffectRotation = "#JUDGELINE_EFFECT_ROTATION"
    JudgelineEffectDirection = "#JUDGELINE_EFFECT_DIRECTION"
    JudgelineEffectAlpha = "#JUDGELINE_EFFECT_ALPHA"
    Slot = "#SLOT"
    SlotPosition = "#SLOT_POSITION"
    SlotSize = "#SLOT_SIZE"
    SlotRotation = "#SLOT_ROTATION"
    SlotDirection = "#SLOT_DIRECTION"
    SlotAlpha = "#SLOT_ALPHA"
    SlotAnimation = "#SLOT_ANIMATION"
    SlotEffect = "#SLOT_EFFECT"
    SlotEffectPosition = "#SLOT_EFFECT_POSITION"
    SlotEffectSize = "#SLOT_EFFECT_SIZE"
    SlotEffectRotation = "#SLOT_EFFECT_ROTATION"
    SlotEffectDirection = "#SLOT_EFFECT_DIRECTION"
    SlotEffectAlpha = "#SLOT_EFFECT_ALPHA"
    Note = "#NOTE"
    NoteSpeed = "#NOTE_SPEED"
    NoteSpeedRandom = "#NOTE_SPEED_RANDOM"
    NotePosition = "#NOTE_POSITION"
    NoteSize = "#NOTE_SIZE"
    NoteRotation = "#NOTE_ROTATION"
    NoteDirection = "#NOTE_DIRECTION"
    NoteColor = "#NOTE_COLOR"
    NoteAlpha = "#NOTE_ALPHA"
    NoteAnimation = "#NOTE_ANIMATION"
    NoteEffect = "#NOTE_EFFECT"
    NoteEffectPosition = "#NOTE_EFFECT_POSITION"
    NoteEffectSize = "#NOTE_EFFECT_SIZE"
    NoteEffectRotation = "#NOTE_EFFECT_ROTATION"
    NoteEffectDirection = "#NOTE_EFFECT_DIRECTION"
    NoteEffectColor = "#NOTE_EFFECT_COLOR"
    NoteEffectAlpha = "#NOTE_EFFECT_ALPHA"
    Marker = "#MARKER"
    MarkerPosition = "#MARKER_POSITION"
    MarkerSize = "#MARKER_SIZE"
    MarkerRotation = "#MARKER_ROTATION"
    MarkerDirection = "#MARKER_DIRECTION"
    MarkerColor = "#MARKER_COLOR"
    MarkerAlpha = "#MARKER_ALPHA"
    MarkerAnimation = "#MARKER_ANIMATION"
    Connector = "#CONNECTOR"
    ConnectorPosition = "#CONNECTOR_POSITION"
    ConnectorSize = "#CONNECTOR_SIZE"
    ConnectorRotation = "#CONNECTOR_ROTATION"
    ConnectorDirection = "#CONNECTOR_DIRECTION"
    ConnectorColor = "#CONNECTOR_COLOR"
    ConnectorAlpha = "#CONNECTOR_ALPHA"
    ConnectorAnimation = "#CONNECTOR_ANIMATION"
    Simline = "#SIMLINE"
    SimlinePosition = "#SIMLINE_POSITION"
    SimlineSize = "#SIMLINE_SIZE"
    SimlineRotation = "#SIMLINE_ROTATION"
    SimlineDirection = "#SIMLINE_DIRECTION"
    SimlineColor = "#SIMLINE_COLOR"
    SimlineAlpha = "#SIMLINE_ALPHA"
    SimlineAnimation = "#SIMLINE_ANIMATION"
    None_ = "#NONE"
    Any = "#ANY"
    All = "#ALL"
    Others = "#OTHERS"
    Short = "#SHORT"
    Long = "#LONG"
    High = "#HIGH"
    Mid = "#MID"
    Low = "#LOW"
    Small = "#SMALL"
    Medium = "#MEDIUM"
    Large = "#LARGE"
    Left = "#LEFT"
    Right = "#RIGHT"
    Up = "#UP"
    Down = "#DOWN"
    Front = "#FRONT"
    Back = "#BACK"
    Center = "#CENTER"
    Top = "#TOP"
    Bottom = "#BOTTOM"
    TopLeft = "#TOP_LEFT"
    TopCenter = "#TOP_CENTER"
    TopRight = "#TOP_RIGHT"
    CenterLeft = "#CENTER_LEFT"
    CenterRight = "#CENTER_RIGHT"
    BottomLeft = "#BOTTOM_LEFT"
    BottomCenter = "#BOTTOM_CENTER"
    BottomRight = "#BOTTOM_RIGHT"
    Clockwise = "#CLOCKWISE"
    Counterclockwise = "#COUNTERCLOCKWISE"
    Forward = "#FORWARD"
    Backward = "#BACKWARD"
    Default = "#DEFAULT"
    Neutral = "#NEUTRAL"
    Red = "#RED"
    Green = "#GREEN"
    Blue = "#BLUE"
    Yellow = "#YELLOW"
    Purple = "#PURPLE"
    Cyan = "#CYAN"
    Simple = "#SIMPLE"
    Easy = "#EASY"
    Normal = "#NORMAL"
    Hard = "#HARD"
    Expert = "#EXPERT"
    Master = "#MASTER"
    Pro = "#PRO"
    Technical = "#TECHNICAL"
    Special = "#SPECIAL"
    KeywordsPlaceholder = "#KEYWORDS_PLACEHOLDER"
    NamePlaceholder = "#NAME_PLACEHOLDER"
    RatingPlaceholder = "#RATING_PLACEHOLDER"
    RatingMinimumPlaceholder = "#RATING_MINIMUM_PLACEHOLDER"
    RatingMaximumPlaceholder = "#RATING_MAXIMUM_PLACEHOLDER"
    TitlePlaceholder = "#TITLE_PLACEHOLDER"
    SubtitlePlaceholder = "#SUBTITLE_PLACEHOLDER"
    ArtistsPlaceholder = "#ARTISTS_PLACEHOLDER"
    TimePlaceholder = "#TIME_PLACEHOLDER"
    AuthorPlaceholder = "#AUTHOR_PLACEHOLDER"
    GenrePlaceholder = "#GENRE_PLACEHOLDER"
    TypePlaceholder = "#TYPE_PLACEHOLDER"
    CategoryPlaceholder = "#CATEGORY_PLACEHOLDER"
    LanguagePlaceholder = "#LANGUAGE_PLACEHOLDER"
    DifficultyPlaceholder = "#DIFFICULTY_PLACEHOLDER"
    LengthPlaceholder = "#LENGTH_PLACEHOLDER"
    LengthMinimumPlaceholder = "#LENGTH_MINIMUM_PLACEHOLDER"
    LengthMaximumPlaceholder = "#LENGTH_MAXIMUM_PLACEHOLDER"
    AdditionalInformationPlaceholder = "#ADDITIONAL_INFORMATION_PLACEHOLDER"
    TimezonePlaceholder = "#TIMEZONE_PLACEHOLDER"
    RegionPlaceholder = "#REGION_PLACEHOLDER"
    ContentPlaceholder = "#CONTENT_PLACEHOLDER"
    CommentPlaceholder = "#COMMENT_PLACEHOLDER"
    ReviewPlaceholder = "#REVIEW_PLACEHOLDER"
    ReplyPlaceholder = "#REPLY_PLACEHOLDER"
    MessagePlaceholder = "#MESSAGE_PLACEHOLDER"
    RolePlaceholder = "#ROLE_PLACEHOLDER"
    PermissionPlaceholder = "#PERMISSION_PLACEHOLDER"
    UserPlaceholder = "#USER_PLACEHOLDER"
    PercentageUnit = "#PERCENTAGE_UNIT"
    SecondUnit = "#SECOND_UNIT"
    MillisecondUnit = "#MILLISECOND_UNIT"
    Tap = "#TAP"
    TapHold = "#TAP_HOLD"
    TapRelease = "#TAP_RELEASE"
    TapFlick = "#TAP_FLICK"
    TapSlide = "#TAP_SLIDE"
    Hold = "#HOLD"
    HoldSlide = "#HOLD_SLIDE"
    HoldFollow = "#HOLD_FOLLOW"
    Release = "#RELEASE"
    Flick = "#FLICK"
    Slide = "#SLIDE"
    SlideFlick = "#SLIDE_FLICK"
    Avoid = "#AVOID"
    Jiggle = "#JIGGLE"
    Newest = "#NEWEST"
    Oldest = "#OLDEST"
    Recommended = "#RECOMMENDED"
    Popular = "#POPULAR"
    Featured = "#FEATURED"
    Competitive = "#COMPETITIVE"
    Holiday = "#HOLIDAY"
    Limited = "#LIMITED"
    Following = "#FOLLOWING"
    Announcement = "#ANNOUNCEMENT"
    Information = "#INFORMATION"
    Help = "#HELP"
    Maintenance = "#MAINTENANCE"
    Event = "#EVENT"
    Update = "#UPDATE"
    Search = "#SEARCH"
    Advanced = "#ADVANCED"
    Related = "#RELATED"
    SameAuthor = "#SAME_AUTHOR"
    SameArtists = "#SAME_ARTISTS"
    SameRating = "#SAME_RATING"
    SameCategory = "#SAME_CATEGORY"
    SameDifficulty = "#SAME_DIFFICULTY"
    SameGenre = "#SAME_GENRE"
    SameVersion = "#SAME_VERSION"
    OtherAuthors = "#OTHER_AUTHORS"
    OtherArtists = "#OTHER_ARTISTS"
    OtherRatings = "#OTHER_RATINGS"
    OtherCategories = "#OTHER_CATEGORIES"
    OtherDifficulties = "#OTHER_DIFFICULTIES"
    OtherGenres = "#OTHER_GENRES"
    OtherVersions = "#OTHER_VERSIONS"
    Draft = "#DRAFT"
    Reviewing = "#REVIEWING"
    Published = "#PUBLISHED"
    Verifying = "#VERIFYING"
    Verified = "#VERIFIED"
    Removed = "#REMOVED"
    Public = "#PUBLIC"
    Private = "#PRIVATE"
    Locked = "#LOCKED"
    Pop = "#POP"
    Rock = "#ROCK"
    Hiphop = "#HIPHOP"
    Country = "#COUNTRY"
    Electronic = "#ELECTRONIC"
    Metal = "#METAL"
    Classical = "#CLASSICAL"
    Folk = "#FOLK"
    Indie = "#INDIE"
    Anime = "#ANIME"
    Vocaloid = "#VOCALOID"
    Remix = "#REMIX"
    Instrumental = "#INSTRUMENTAL"
    ShortVersion = "#SHORT_VERSION"
    LongVersion = "#LONG_VERSION"
    LiveVersion = "#LIVE_VERSION"
    Report = "#REPORT"
    Reason = "#REASON"
    IllegalActivities = "#ILLEGAL_ACTIVITIES"
    Cheating = "#CHEATING"
    Afk = "#AFK"
    Spamming = "#SPAMMING"
    VerbalAbuse = "#VERBAL_ABUSE"
    InappropriateLanguage = "#INAPPROPRIATE_LANGUAGE"
    NegativeAttitude = "#NEGATIVE_ATTITUDE"
    Dnf = "#DNF"
    Suggestions = "#SUGGESTIONS"
    SuggestionsPerPlayer = "#SUGGESTIONS_PER_PLAYER"
    MatchScoring = "#MATCH_SCORING"
    MatchTiebreaker = "#MATCH_TIEBREAKER"
    MatchCount = "#MATCH_COUNT"
    MatchLimit = "#MATCH_LIMIT"
    RoundScoring = "#ROUND_SCORING"
    RoundTiebreaker = "#ROUND_TIEBREAKER"
    RoundCount = "#ROUND_COUNT"
    RoundLimit = "#ROUND_LIMIT"
    TeamScoring = "#TEAM_SCORING"
    TeamTiebreaker = "#TEAM_TIEBREAKER"
    TeamCount = "#TEAM_COUNT"
    TeamLimit = "#TEAM_LIMIT"
    Qualified = "#QUALIFIED"
    Disqualified = "#DISQUALIFIED"
    Ranking = "#RANKING"
    Score = "#SCORE"
    Player = "#PLAYER"
    Spectator = "#SPECTATOR"
    Referee = "#REFEREE"
    Eliminated = "#ELIMINATED"
    Finalist = "#FINALIST"
    Finished = "#FINISHED"
    Winner = "#WINNER"
    GoldMedal = "#GOLD_MEDAL"
    SilverMedal = "#SILVER_MEDAL"
    BronzeMedal = "#BRONZE_MEDAL"
    Team1 = "#TEAM_1"
    Team2 = "#TEAM_2"
    Team3 = "#TEAM_3"
    Team4 = "#TEAM_4"
    Team5 = "#TEAM_5"
    Team6 = "#TEAM_6"
    Team7 = "#TEAM_7"
    Team8 = "#TEAM_8"
    TeamRed = "#TEAM_RED"
    TeamGreen = "#TEAM_GREEN"
    TeamBlue = "#TEAM_BLUE"
    TeamYellow = "#TEAM_YELLOW"
    TeamPurple = "#TEAM_PURPLE"
    TeamCyan = "#TEAM_CYAN"
    TeamWhite = "#TEAM_WHITE"
    TeamBlack = "#TEAM_BLACK"
    Upload = "#UPLOAD"
    Submit = "#SUBMIT"
    Edit = "#EDIT"
    Like = "#LIKE"
    Liked = "#LIKED"
    Dislike = "#DISLIKE"
    Disliked = "#DISLIKED"
    Bookmark = "#BOOKMARK"
    Bookmarked = "#BOOKMARKED"
    Delete = "#DELETE"
    Restore = "#RESTORE"
    Confirm = "#CONFIRM"
    Cancel = "#CANCEL"
    Increase = "#INCREASE"
    Decrease = "#DECREASE"
    Upvote = "#UPVOTE"
    Upvoted = "#UPVOTED"
    Downvote = "#DOWNVOTE"
    Downvoted = "#DOWNVOTED"
    Agree = "#AGREE"
    Agreed = "#AGREED"
    Disagree = "#DISAGREE"
    Disagreed = "#DISAGREED"
    Lock = "#LOCK"
    Unlock = "#UNLOCK"
    Pin = "#PIN"
    Unpin = "#UNPIN"
    Follow = "#FOLLOW"
    Unfollow = "#UNFOLLOW"
    Subscribe = "#SUBSCRIBE"
    Unsubscribe = "#UNSUBSCRIBE"
    Publish = "#PUBLISH"
    Unpublish = "#UNPUBLISH"
    Show = "#SHOW"
    Hide = "#HIDE"
    Allow = "#ALLOW"
    Disallow = "#DISALLOW"
    Approve = "#APPROVE"
    Deny = "#DENY"
    Accept = "#ACCEPT"
    Reject = "#REJECT"
    Star = "#STAR"
    Starred = "#STARRED"
